#include "rizeni.h"
#include "nucleo_qtr.h"

volatile uint32_t _ticks = 0;

bool btnOn=false;
uint32_t tstamp;
ePodvozky aktivniPodvozek;

void SysTick_Handler(void)
{
  _ticks++;
}

//promenne pro DemoCtverec
int sl;
int sr;
int krok;
int strany;
uint8_t ctverecMaxSpeed;
uint32_t ctverecDobaOtocky;
uint8_t ctverecRychlostOtocky;

void DemoCtverec()
{
  switch(aktivniPodvozek)               //nastaveni parametru podle podvozku
  {
    case lego:
      ctverecMaxSpeed = 100;
      ctverecDobaOtocky = 580;
      ctverecRychlostOtocky = 80;
      break;
    case zumo:
      ctverecMaxSpeed = 100;
      ctverecDobaOtocky = 520;
      ctverecRychlostOtocky = 80;
      break;
    case pasak:
      ctverecMaxSpeed = 255;
      ctverecDobaOtocky = 850;
      ctverecRychlostOtocky = 255;
      break;
    case kolovyPodvozek:
      ctverecMaxSpeed = 150;
      ctverecDobaOtocky = 640;
      ctverecRychlostOtocky = 100;
      break;
  }
  switch(krok)
  {
    case 0:                             //zrychlovani
      if(sl<ctverecMaxSpeed)            //rychlost ktere chci dosahnout
      {
        if(_ticks>tstamp)               //kazdych 10ms pridam o 1
        {
          tstamp=_ticks+10;
          sl++;
          sr++;
        }
      }
      else                             //po dosazeni pozadovane preskocim na dalsi krok
      {
        krok++;
        tstamp = _ticks;
      }
      break;
    case 1:                             //drzeni rychlosti
      if(_ticks>=tstamp+1000)           //cas po ktery drzime rychlost
      {
        krok++;                         //po uplynuti preskocim na dalsi krok
      }
      break;
    case 2:                            //zpomalovani
      if(sl>0)                         //dokud nezastavim do nuly
      {
        if(_ticks>tstamp)              //kazdych 10ms zpomalim o 1
        {
          tstamp=_ticks+10;
          sl--;
          sr--;
        }
      }
      else                            //dalsi krok
      {
        krok++;
      }
      break;

    case 3:                            //otoceni o 90st
      if(_ticks>tstamp)
      {
        tstamp=_ticks+ctverecDobaOtocky; //doba otaceni
        if((sl==0) && (sr==0))
        {
          sl = ctverecRychlostOtocky;    //rychlost otaceni
          sr = -ctverecRychlostOtocky;
        }
        else                           //pak zastavit
        {
          sl=0;
          sr=0;
          strany++;                     //pricist stranu
          krok++;                       //dalsi krok
        }
      }
      break;
    case 4:                             //vyhodnoceni
      if(strany>=4)                     //uz je cely ctverec
      {
        strany = 0;                     //ano, resetuju promenne
        btnOn = false;
        krok = 0;
      }
      else                             //ne, vracim se na zacatek
      {
        krok = 0;
      }
      break;
  }
  if(aktivniPodvozek == kolovyPodvozek)        //korekce pro kolovy podvozek
    MotorsSetSpeed(sl,sr+12);
  else
    MotorsSetSpeed(sl,sr);              //nastaveni motoru na pozadovanou rychlost
}

volatile gpio_pin legoSenzorPins[LEGO_NUM_SENZORS]=   //pole senzoru lego
{
    { GPIOB, 1 },
    { GPIOC, 4 },
    { GPIOA, 7 },
    { GPIOC, 5 },
    { GPIOB, 0 },
    { GPIOC, 0 },
    { GPIOC, 3 },
    { GPIOC, 2 },
};

volatile gpio_pin zumoSenzorPins[ZUMO_NUM_SENZORS]=   //pole senzoru zumo
{
    {GPIOB,5},
    {GPIOB,0},
    {GPIOA,7},
    {GPIOA,0},
    {GPIOA,4},
    {GPIOB,4},
};

//promenne pro line follower
int motorSpeed = 0;
float error = 0;
float proportional = 0;
float integral = 0;
float derivative = 0;
float lastError = 0;
bool calib = false;
bool init = false;
float kp;
float ki;
float kd;
int baseSpeed;
int senzorsNum;

void DemoLineFollower()
{
  if(init == false)               //pokud neni inicializace
  {
    switch (aktivniPodvozek)      //nastaveni parametru podle podvozku
    {
      case lego:
        QTR_InitSensors(legoSenzorPins, LEGO_NUM_SENZORS, false); //inicializace senzoru
        kp = (1/3.0);             //proporcionalni konstanta
        ki = 0;                   //integracni konstanta
        kd = 40.0;                //derivacni konstanta
        baseSpeed = 90;           //zakladni rychlost
        senzorsNum = LEGO_NUM_SENZORS;  //pocet senzoru
        break;
      case zumo:
        QTR_InitSensors(zumoSenzorPins, ZUMO_NUM_SENZORS, true);  //inicializace senzoru
        kp = (1/1.0);             //proporcionalni konstanta
        ki = 0*(1/70.0);          //integracni konstanta
        kd = 40.0;                //derivacni konstanta
        baseSpeed = 100;          //zakladni rychlost
        senzorsNum = ZUMO_NUM_SENZORS;  //pocet senzoru
        break;
      case kolovyPodvozek:
      case pasak:
      default:
        MotorsSetSpeed(0,0);
        break;
    }
    QTR_EnableMeassure(true);      //povoeni mereni
    init = true;                   //nastavime priznak inicializace
  }
  if(calib == false)               //pokud neni zkalibrovani
  {
    for(int i = 0;i<1000;i++)      //postup kalibrace
    {
      if((i<250)||(i>750))         //otaceni vlevo, vpravo a zpet do stredu
      {
        MotorsSetSpeed(60,-60);
      }
      else
      {
        MotorsSetSpeed(-60,60);
      }
      QTR_StepCalibrate();          //kalibrujeme senzory
      tstamp = _ticks+5;            //cekame 5 ms v kazdem kroku
      while(_ticks<tstamp)
        ;
    }
    MotorsSetSpeed(0,0);            //zastavime
    calib = true;                   //priznak kalibrace
    btnOn = false;                  //promenna od tlacitka se resetuje
  }else                             //pokud je zkalibrovano
  {
    if(aktivniPodvozek == zumo)
      error = ((QTR_LineDetect()-2500.0)/senzorsNum); //prepocitame chybu ze senzoru na +-1000
    else
      error = ((QTR_LineDetect()- 4000.0)/senzorsNum);   //prepocitame chybu ze senzoru na +-1000

    proportional = error;           //proporcionalni slozka = chyba
    integral +=error;               //integracni slozka
    derivative = lastError - error; //derivacni slozka
    lastError = error;              //posledni chyba

    if(integral > 1000)             //omezeni integracni slozky
      integral = 1000;
    if(integral<-1000)
      integral = -1000;

    motorSpeed = (kp*proportional)+(ki*integral)+(kd*derivative); //vypocet aditivni rychlosti

    sl = baseSpeed+motorSpeed;                                    //nastaveni L rychlosti
    sr = baseSpeed-motorSpeed;                                    //nastaveni R rychlosti

    if(sl<0)        //omezeni proti otaceni smerem vzad
      sl = 0;
    if(sr<0)
      sr = 0;

    if(sl>254)      //omezeni max hodnoty
      sl = 254;
    if(sr>254)
      sr = 254;
    MotorsSetSpeed(sl,sr);    //nastaveni rychlosti motoru
  }
}

int main(void)
{
  SetClock100MHz(clockSourceHSE);     //nastaveni hodin na 100 Mhz
  SystemCoreClockUpdate();            //update hodin
  Nucleo_SetPinGPIO(BOARD_BTN_BLUE,ioPortInputPU);  //modre tlacitko na nucleo jako vstup
  uint32_t tbtn = 0;                  //timestamp pro tlacitko
  bool btn;                           //promenne pro tlacitka
  bool btnLast;
  SysTick_Config(SystemCoreClock / 1000); //nastaveni SysTicku

  //aktivniPodvozek = MotorsInit(lego);   //volba podvozku
  //aktivniPodvozek = MotorsInit(zumo);
  aktivniPodvozek = MotorsInit(kolovyPodvozek);
  //aktivniPodvozek = MotorsInit(pasak);
  while (1)
  {
    if(_ticks >= tbtn)            //cteni tlacitka kazdych 5ms
    {
      tbtn = _ticks + 5;
      if(aktivniPodvozek == zumo)
        btn = GPIORead(ZUMO_BUTTON);
      else
        btn = GPIORead(BOARD_BTN_BLUE);
      if(btn != btnLast)
      {
        if(!btn)
        {
          btnOn = !btnOn;
        }
        btnLast = btn;
      }
    }

    if(btnOn)
    {
        DemoCtverec();          //volba demo programu
      //DemoLineFollower();
    }
    else
    {
      MotorsSetSpeed(0,0);
    }

  }

}
