/*
 * nucleo_qtr.c
 *
 *  Created on: 11. 2. 2018
 *      Author: weissar
 */

#include "nucleo_qtr.h"

#define MAX_NUM_SENZORS   10

static volatile uint16_t senzors[MAX_NUM_SENZORS];
static volatile uint16_t senzorsFinal[MAX_NUM_SENZORS];

static uint16_t sensMax[MAX_NUM_SENZORS];
static uint16_t sensMin[MAX_NUM_SENZORS];

static float fSensK[MAX_NUM_SENZORS];
static uint16_t uSensO[MAX_NUM_SENZORS];

static volatile bool meassureProgress = false;
static volatile bool meassureEnable = false;
static volatile int cntMeassureStart = 0;
static volatile int cntMeassurePulse = 0;

static volatile gpio_pin *_pins = NULL;
static volatile int _pinCount = 0;

static volatile bool _useRC = true;
static volatile bool _useRegularAD = true;
static volatile int _cntRegular = 0;

#define AD_CHANNELS_COUNT   16

static gpio_pin adchannels[AD_CHANNELS_COUNT] =
{
    { GPIOA, 0 },
    { GPIOA, 1 },
    { NULL, 0 },      // not on Nucleo-64
    { NULL, 0 },      // not on Nucleo-64
    { GPIOA, 4 },
    { GPIOA, 5 },
    { GPIOA, 6 },
    { GPIOA, 7 },

    { GPIOB, 0 },
    { GPIOB, 1 },
    { GPIOC, 0 },
    { GPIOC, 1 },
    { GPIOC, 2 },
    { GPIOC, 3 },
    { GPIOC, 4 },
    { GPIOC, 5 }
};

bool QTR_InitSensors(volatile gpio_pin *pins, int count, bool useRC)
{
  if ((pins == NULL) || (count == 0) || (count > MAX_NUM_SENZORS))
  {
    _pinCount = 0;     // priznak spatne inicializace
    return false;
  }
  else
  {
    _pinCount = count;
    _pins = pins;      //TODO make local copy

    _useRC = useRC;
  }

  for(int i = 0; i < _pinCount; i++)
  {
    if (_useRC)
    {
      Nucleo_SetPinGPIO(_pins[i].gpio, _pins[i].pin, ioPortOutputPP);
      GPIOWrite(_pins[i].gpio, _pins[i].pin, true);
    }
    else
    {
      Nucleo_SetPinGPIO(_pins[i].gpio, _pins[i].pin, ioPortAnalog);
    }

    senzors[i] = 0;
    senzorsFinal[i] = 0;
  }

  QTR_ClearCalibrate();

  if (_useRC)
  {
    // AD settings not required
  }
  else
  {
    if (_pinCount > 4)
    {
      _useRegularAD = true;
      _cntRegular = 0;
    }
    else
      _useRegularAD = false;

    if (!(RCC->APB2ENR & RCC_APB2ENR_ADC1EN))
    {
      RCC->APB2ENR |= RCC_APB2ENR_ADC1EN;
      RCC->APB2RSTR |= RCC_APB2RSTR_ADCRST;
      RCC->APB2RSTR &= ~RCC_APB2RSTR_ADCRST;
    }

    ADC1->CR1 = 0   // RES = 00 = 12b
        | ADC_CR1_SCAN;
    ADC1->CR2 = 0   // ALIGN = 0 = right align
        ;

    ADC->CCR = 0              // ADCPRE = 00 = APB2 / 2
        | ADC_CCR_ADCPRE_1 | ADC_CCR_ADCPRE_0    // ADCPRE = 11 = APB2 / 8 - jedeme na 100MHz !!
        ;
    // doba - 16M/2 za ADCPRE - 8MHz, 125ns, 28 + 12 = 40 taktu = 5us, 56+12=68=8.5us

    ADC1->SMPR1 = 0;
    ADC1->SMPR2 = 0;

    if (_useRegularAD)
    {
      ADC1->CR2 |= ADC_CR2_EOCS;      // IRQ for every conversion
      ADC1->JSQR = 0;
      ADC1->SQR1 = (_pinCount - 1) << 20;   // bity L, 0000 = 1 prevod, 0001 = 2 prevody, ...
      ADC1->SQR2 =  ADC1->SQR3 = 0;
    }
    else
    {
      ADC1->JSQR = (_pinCount - 1) << 20;   // bity JL, 00 = 1 prevod, 11 = 4 prevody
      ADC1->SQR1 =  ADC1->SQR2 =  ADC1->SQR3 = 0;
    }

    for(int i = 0; i < _pinCount; i++)
    {
      int ch = -1;
      for(int x = 0; x < AD_CHANNELS_COUNT; x++)
      {
        if ((_pins[i].gpio == adchannels[x].gpio) && (_pins[i].pin == adchannels[x].pin))
        {
          ch = x;
          break;
        }
      }

      if (ch < 0)             // not found ?
      {
        _pinCount = 0;
        return false;
      }

      if (ch < 10)    // 0 .. 9
        ADC1->SMPR1 |= 3 << (3 * ch);
      else
        ADC1->SMPR2 |= 3 << (3 * (ch - 10));       // 011 = 56 cycles
      // doba - 16M/2 za ADCPRE - 8MHz, 125ns, 28 + 12 = 40 taktu = 5us, 56+12=68=8.5us

      if (_useRegularAD)
      {
        if (i >= 13)
          ADC1->SQR1 |= ch << (5 * (i - 13));    // SQ16 .. SQ13
        else
          if (i >= 7)
            ADC1->SQR2 |= ch << (5 * (i - 7));    // SQ12 .. SQ7
          else
            ADC1->SQR3 |= ch << (5 * i);    // SQ6 .. SQ1
      }
      else
        ADC1->JSQR |= ch << (5 * (3 - i));      // 5 bits/channel, prvni JSQ4
    }

    if (_useRegularAD)
      ADC1->CR1 |= ADC_CR1_EOCIE; //BB_REG(ADC1->CR1, 5) = 1;     // enable EOC IRQ
    else
      ADC1->CR1 |= ADC_CR1_JEOCIE; //BB_REG(ADC1->CR1, 7) = 1;    // enable JEOC IRQ
    NVIC_EnableIRQ(ADC_IRQn);     // pozor, spolecne pro ADC1,2,3

    ADC1->CR2 |= ADC_CR2_ADON; //BB_REG(ADC1->CR2, 0) = 1;        // on
  }

  // use TIM5 for timing
  if (!(RCC->APB1ENR & RCC_APB1ENR_TIM5EN))
  {
    RCC->APB1ENR |= RCC_APB1ENR_TIM5EN;
    RCC->APB1RSTR |= RCC_APB1RSTR_TIM5RST;
    RCC->APB1RSTR &= ~RCC_APB1RSTR_TIM5RST;
  }

  TIM5->CR1 = 0
      | TIM_CR1_DIR       // down
      ;
  TIM5->CR2 = 0;

  TIM5->PSC = GetTimerClock(5) / 1E6 - 1;       // 1us

  if (_useRC)
  {
    //TODO zmena z 5 na 10
    TIM5->ARR = 10-1;        // 5x = 5us
  }
  else
  {
    TIM5->ARR = 200-1;      // 100x = 100us
  }

  TIM5->DIER = TIM_DIER_UIE;
  NVIC_EnableIRQ(TIM5_IRQn);

  TIM5->CR1 |= TIM_CR1_CEN;

  return true;
}

bool QTR_ClearCalibrate(void)
{
  if (_pinCount == 0)
    return false;

  for(int i = 0; i < _pinCount; i++)
  {
    sensMax[i] = 0;
    sensMin[i] = UINT16_MAX;
    uSensO[i] = 0;
    fSensK[i] = 1;
  }
  return true;
}

bool QTR_StepCalibrate(void)
{
  if (_pinCount == 0)
    return false;

  for(int i = 0; i < _pinCount; i++)
  {
    if (sensMax[i] < senzorsFinal[i])
      sensMax[i] = senzorsFinal[i];
    if (sensMin[i] > senzorsFinal[i])
      sensMin[i] = senzorsFinal[i];
  }

  for(int i = 0; i < _pinCount; i++)
  {
    uSensO[i] = sensMin[i];
    fSensK[i] = ((float)1000) / (sensMax[i] - sensMin[i]);
  }
  for(int i = 0; i < 8; i++)
          printf("%d: %d - %d  O=%d K=%f\n", i, sensMin[i], sensMax[i], uSensO[i],fSensK[i]);
  return true;
  //zobrazeni vysledku kalibrace

}





uint16_t QTR_LineDetect(void)
{
  uint32_t _sumW = 0;
  uint32_t _sum = 0;
  uint32_t _moc = 0;
  bool _lineDetected = false;       // tyto 2 veci inspirovane QTR lib pro Arduino
  static uint16_t last_val = 0;

  if (_pinCount == 0)
    return 0;

  for(int i = 0; i < _pinCount; i++)
  {
    uint16_t val = (uint16_t)((senzorsFinal[i] - uSensO[i]) * fSensK[i]);

    if (val > 200)
      _lineDetected = true;  // aspon nekde je trochu cerna

    if (val > 50)
    {
      _sumW += _moc * val;
      _sum += val;
    }

    _moc += 1000;
  }

  if (_lineDetected)
  {
    last_val = _sumW / _sum;
  }
  else
  {
    if (last_val < ((_pinCount - 1) * 1000 / 2))
      last_val = 0;
    else
      last_val = (_pinCount - 1) * 1000;
  }

  //  printf(" = %d", last_val);
  //  puts("");
  return last_val;
}

bool QTR_FillSensors(uint16_t *pSens)
{
  if (_pinCount == 0)
    return false;

  for(int i =0; i < _pinCount; i++)
    pSens[i] = (uint16_t)((senzorsFinal[i] - uSensO[i]) * fSensK[i]);

  return true;
}

bool QTR_FillPureSensVals(uint16_t *pSens)
{
  if (_pinCount == 0)
    return false;

  for(int i =0; i < _pinCount; i++)
    pSens[i] = senzorsFinal[i];

  return true;
}

void QTR_EnableMeassure(bool en)
{
  meassureEnable = en;
}

void TIM5_IRQHandler(void)
{
  if (TIM5->SR & TIM_SR_UIF)
  {
    TIM5->SR &= ~TIM_SR_UIF; //BB_REG(TIM5->SR, 0) = 0;        // clear UIF

    if (!meassureEnable)
    {
      cntMeassureStart = 0;
      meassureProgress = false;

      return;     // nothing to do
    }

    if (_useRC)     // mereni casu
    {
      if (meassureProgress)         // probiha mereni casu ?
      {
        //TODO zmena z 4000 na 2000
        if (cntMeassurePulse >= 2000)    // 4000x 5us = 20ms
        {
          for(int i = 0; i < _pinCount; i++)
            senzorsFinal[i] = senzors[i];    // double buffer

          meassureProgress = false;  // a je domereno
          cntMeassureStart = 0;
        }
        else
        {
          cntMeassurePulse++;

          for(int i = 0; i < _pinCount; i++)  // pocitej dobu v log. 1
          {
            if (GPIORead(_pins[i].gpio, _pins[i].pin))
              senzors[i]++;
          }
        }
      }
      else                          // ne, je to uvodni puls
      {
        if (cntMeassureStart == 0)  // zaciname ?
        {
          //          GPIOB->MODER |= 0x00000001 << (2 * 5);    // 01 = output a OTYPER je 0 = PP

          for(int i = 0; i < _pinCount; i++)  // pocitej dobu v log. 1
            _pins[i].gpio->MODER |= 0x00000001 << (2 * _pins[i].pin);

          for(int i = 0; i < _pinCount; i++)  // pocitej dobu v log. 1
          {
            GPIOWrite(_pins[i].gpio, _pins[i].pin, true);
            senzors[i] = 0;
          }
        }
        //TODO zmena
        if (cntMeassureStart >= 1)      // 2x 5us = 10us
        {
          //          GPIOB->MODER &= ~(0x00000003 << (2 * 5));   // 00 = input (float)

          for(int i = 0; i < _pinCount; i++)  // pocitej dobu v log. 1
            _pins[i].gpio->MODER &= ~(0x00000003 << (2 * _pins[i].pin));

          cntMeassurePulse = 0;
          meassureProgress = true;  // a ted se meri preklopeni pri vybijeni C
        }
        else                        // tak merime uvodni nabijeci puls
          cntMeassureStart++;
      }
    }
    else
    {
      if (_useRegularAD)
      {
        _cntRegular = 0;
        ADC1->CR2 |= ADC_CR2_SWSTART;   //BB_REG(ADC1->CR2, 30) = 1;    // ADC_CR2_SWSTART
      }
      else
      ADC1->CR2 |= ADC_CR2_JSWSTART;    //BB_REG(ADC1->CR2, 22) = 1;    // ADC_CR2_JSWSTART
    }
  }
}

void ADC_IRQHandler(void)
{
  //TODO test Event source/type
  if (_useRegularAD)
  {
    ADC1->SR &= ~ADC_SR_EOC;  //BB_REG(ADC1->SR, 1) = 0;   // bit 1 -  ADC1->SR &= ~ADC_SR_EOC;

    if (_cntRegular < MAX_NUM_SENZORS)
    {
      senzorsFinal[_cntRegular] = ADC1->DR;
      _cntRegular++;
      //TODO if last position read, set flag "finished all"
    }
    else
      (void)ADC1->DR;     // dummy read
  }
  else
  {
    ADC1->SR &= ~ADC_SR_JEOC; //BB_REG(ADC1->SR, 2) = 0;   // bit 2 -  ADC1->SR &= ~ADC_SR_JEOC;   // no regular - e.g. Injection

    if (_pinCount > 0) senzorsFinal[0] = ADC1->JDR1;
    if (_pinCount > 1) senzorsFinal[1] = ADC1->JDR2;
    if (_pinCount > 2) senzorsFinal[2] = ADC1->JDR3;
    if (_pinCount > 3) senzorsFinal[3] = ADC1->JDR4;
  }
}
